#!/usr/bin/env python3
import os
import errno
import json
import getpass
import argparse
import click

from . import _config
from . import track_cli
from ._version import __version__


CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])

class LowerCaseFormatter(click.HelpFormatter):
    def write_usage(self, prog, args='', prefix='usage: '):
        super(LowerCaseFormatter, self).write_usage(prog, args, prefix)
click.Context.formatter_class = LowerCaseFormatter

@click.group(context_settings=CONTEXT_SETTINGS,
             options_metavar='[-h] [-V]',
             subcommand_metavar='<command> [<args>]')
@click.version_option(__version__, '-V', '--version')
@click.pass_context
def cli(ctx):
    """
    Track completion of your jobs!
    """
    pass 


@cli.command(options_metavar='[-h]')
def config():
    """
    Configure user details and auth for notifications
    """
    config_data = _config.get_config_data()
    input_config_data = {
        "machine": input('What name should this device go by? ({}) '
                            .format(config_data.get("machine", ""))),
        "email": input('What email to use? ({}) '
                        .format(config_data.get("email", ""))),
        "smtp_server": input('What SMTP server does your email use? ({}) '
                                .format(config_data.get("smtp_server", ""))),
        "email_password": getpass.getpass('Email password? ({}) '
                                            .format(("*"*16 
                                                    if config_data.get("email_password") is not None 
                                                    else ""))),
        "twilio_receive_number": input('What phone number to receive SMS? ({}) '
                                        .format(config_data.get("twilio_receive_number", ""))),
        "twilio_send_number": input('What Twilio number to send SMS? ({}) '
                                    .format(config_data.get("twilio_send_number", ""))),
        "twilio_account_sid": input('Twilio Account SID? ({}) '
                                    .format(config_data.get("twilio_account_sid", ""))),
        "twilio_auth_token": getpass.getpass('Twilio Auth Token? ({}) '
                                                .format(("*"*16 
                                                        if config_data.get("twilio_auth_token") is not None 
                                                        else "")))
    }
    # Only save inputs if they aren't empty
    for key in input_config_data:
        if input_config_data[key] != "":
            config_data[key] = input_config_data[key]
    with open(f"{_config.JORT_DIR}/config", "w") as f:
        json.dump(config_data, f)


@cli.command(options_metavar='[<options>]')
def inspect():
    """
    Get saved job details from database
    """
    pass


@cli.command(short_help='Track <job>, either a shell command or an existing PID',
             no_args_is_help=True,
             options_metavar='[<options>]')
@click.argument('job', nargs=-1, metavar='<job>')
@click.option('-t', '--text', is_flag=True, 
              help='send SMS text at job exit')
@click.option('-e', '--email', is_flag=True, 
              help='send email at job exit')
@click.option('-d', '--database', is_flag=True, 
              help='store job details in database')
@click.option('-s', '--session', metavar='<session>',
              help='job session name for database')
@click.option('-u', '--unique', is_flag=True, 
              help='skip if session & job have completed previously with no errors')
@click.option('-o', '--output', is_flag=True,
              help='save stdout/stderr output when sending email notification')
@click.option('--shell', is_flag=True,
              help='use shell execution when tracking new job')
@click.option('-v', '--verbose', is_flag=True, 
              help='print job payloads and all info')
def track(job, text, email, database, session, unique, output, shell, verbose):
    """
    Track <job>, which is either a shell command or an existing PID
    """
    if len(job) == 1 and job[0].isdigit():
        pid = int(job[0])
        # Use PID tracking
        print(f"Tracking existing process PID at: {pid}")
        track_cli.track_existing(pid,
                                 to_db=database,
                                 session_name=session,
                                 send_text=text,
                                 send_email=email,
                                 verbose=verbose)
    else:
        # Run command and track execution
        joined_command = ' '.join(job)
        print(f"Tracking command `{joined_command}`")
        track_cli.track_new(joined_command,
                            use_shell=shell,
                            store_stdout=output,
                            save_filename=None,
                            to_db=database,
                            session_name=session,
                            unique=unique,
                            send_text=text,
                            send_email=email,
                            verbose=verbose)



def main():
    parser = argparse.ArgumentParser(
        description='Track completion of your jobs!'
    )

    parser.add_argument(
        '-c',
        '--command',
        nargs='+',
        help='full command to track',
    )

    # May potentially support multiple processes
    parser.add_argument(
        '-p',
        '--pid',
        type=int,
        help='PID of existing job to track',
    )

    # Save stdout/stderr output
    parser.add_argument('-o',
                        '--output',
                        action='store_true',
                        help='save stdout/stderr output')

    parser.add_argument('--use-shell',
                        action='store_true',
                        help='use shell execution for tracking new process')

    # Send SMS at job completion
    parser.add_argument('-s',
                        '--sms',
                        action='store_true',
                        help='send SMS at job exit')
    
    # Send email at job completion
    parser.add_argument('-e',
                        '--email',
                        action='store_true',
                        help='send email at job exit')

    parser.add_argument('-d',
                        '--database',
                        action='store_true',
                        help='store job in database')
    
    parser.add_argument(
        '--session',
        type=str,
        help='job session name, for database',
    )

    parser.add_argument('-u',
                        '--unique',
                        action='store_true',
                        help='skip if session+job have completed previously with no errors')
    
    # Init / info
    parser.add_argument('-i',
                        '--init',
                        action='store_true',
                        help='enter information needed for notifications')

    # Verbose
    parser.add_argument('-v',
                        '--verbose',
                        action='store_true',
                        help='print payloads and all info')

    # Version
    parser.add_argument('-V',
                        '--version',
                        action='version',
                        version=f'%(prog)s {__version__}'
                        )

    args = parser.parse_args()

    if args.init:
        config_data = _config.get_config_data()
        input_config_data = {
            "machine": input('What name should this device go by? ({}) '
                                .format(config_data.get("machine", ""))),
            "email": input('What email to use? ({}) '
                            .format(config_data.get("email", ""))),
            "smtp_server": input('What SMTP server does your email use? ({}) '
                                    .format(config_data.get("smtp_server", ""))),
            "email_password": getpass.getpass('Email password? ({}) '
                                                .format(("*"*16 
                                                        if config_data.get("email_password") is not None 
                                                        else ""))),
            "twilio_receive_number": input('What phone number to receive SMS? ({}) '
                                            .format(config_data.get("twilio_receive_number", ""))),
            "twilio_send_number": input('What Twilio number to send SMS? ({}) '
                                        .format(config_data.get("twilio_send_number", ""))),
            "twilio_account_sid": input('Twilio Account SID? ({}) '
                                        .format(config_data.get("twilio_account_sid", ""))),
            "twilio_auth_token": getpass.getpass('Twilio Auth Token? ({}) '
                                                    .format(("*"*16 
                                                            if config_data.get("twilio_auth_token") is not None 
                                                            else "")))
        }
        # Only save inputs if they aren't empty
        for key in input_config_data:
            if input_config_data[key] != "":
                config_data[key] = input_config_data[key]
        with open(f"{_config.JORT_DIR}/config", "w") as f:
            json.dump(config_data, f)            
    if args.command and args.pid:
        parser.error('Please specify only one command or process to track.')
    elif args.command is None and args.pid is None and not args.init:
        parser.print_help()
    elif args.command:
        # # Grab all aws credentials; either from file or interactively
        # aws_credentials = auth.login()
        joined_command = ' '.join(args.command)
        print(f"Tracking command '{joined_command}'")
        track_cli.track_new(joined_command,
                            use_shell=args.use_shell,
                            store_stdout=args.output,
                            save_filename=None,
                            to_db=args.database,
                            session_name=args.session,
                            unique=args.unique,
                            send_text=args.sms,
                            send_email=args.email,
                            verbose=args.verbose)
    elif args.pid:
        # # Grab all aws credentials; either from file or interactively
        # aws_credentials = auth.login()
        print(f"Tracking existing process PID at: {args.pid}")
        track_cli.track_existing(args.pid,
                                 to_db=args.database,
                                 session_name=args.session,
                                 send_text=args.sms,
                                 send_email=args.email,
                                 verbose=args.verbose)
    elif args.init:
        pass
    else:
        parser.error('Something went wrong!')

if __name__ == '__main__':
    cli()


import sys
import time
import sqlite3
import logging
import traceback
import functools
import shortuuid
import contextlib

from . import _config
from . import checkpoint
from . import datetime_utils
        

class Tracker(object):
    """
    A class to time sections of Python scripts by creating and closing timing
    checkpoints. 

    Parameters
    ----------
    logname : str
        Filename for timing logs
    verbose : int, optional
        Options for verbosity. 0 for none, 1 for INFO, and 2 for DEBUG.
    to_db : bool, optional
        Save all checkpoints to database
    session_name : str, optional
        Name of job session, if saving jobs to database

    :ivar date_created: time of initialization
    :ivar machine: name of local machine
    :ivar checkpoints: dict of Checkpoints
    :ivar open_checkpoint_payloads: dict of job status payloads for open Checkpoints
    :ivar logname: log filename
    :iver to_db: option to save all checkpoints to database
    :iver session_name: name of job session
    """
    def __init__(self, logname="tracker.log", verbose=0, to_db=False, session_name=None):
        self.date_created = datetime_utils.get_iso_date()
        self.machine = _config.get_config_data().get("machine")
        self.checkpoints = {}
        self.open_checkpoint_payloads = {}

        # Manage session name, id; if session name is provided, get the id from db
        self.to_db = to_db
        self.session_name = session_name
        self.session_id = shortuuid.uuid()

        with contextlib.closing(sqlite3.connect(f"{_config.JORT_DIR}/jort.db")) as con:
            cur = con.cursor()
            if self.session_name is not None:
                sql = "SELECT session_id FROM sessions WHERE session_name == ?"
                res = cur.execute(sql, (self.session_name,))
                row = res.fetchone()
                if row is not None:
                    self.session_id = row[0]
            else:
                self.session_name = self.session_id
                if self.to_db:
                    sql = (
                        "INSERT INTO sessions VALUES(?, ?)"
                    )
                    cur.execute(sql, (self.session_id, self.session_name))
                    con.commit()

        self.logname = logname
        if verbose != 0:
            print(f"Starting session `{self.session_name}`")
            # if verbose == 1:
            #     level = logging.INFO
            # else:
            #     level = logging.DEBUG
            # file_handler = logging.FileHandler(filename=self.logname, mode="w")
            # stdout_handler = logging.StreamHandler(sys.stdout)
            # handlers = [file_handler, stdout_handler]

            # logging.basicConfig(level=level,
            #                     format="%(asctime)s %(name)-15s %(levelname)-8s %(message)s",
            #                     handlers=handlers, 
            #                     force=True)
        
    def start(self, name=None, date_created=None):
        """
        Open checkpoint and start timer. Creates initial job status payload for use
        with notifications.

        Parameters
        ----------
        name : str
            Checkpoint name
        date_created : str, optional
            For an existing process, instead set this input as the creation date
        """
        if name is None:
            name = "Misc"
        name = str(name)
        if name in self.open_checkpoint_payloads:
            raise RuntimeError(f"Open checkpoint named {name} already exists")
        
        if date_created is not None:
            start = date_created
        else:
            start = datetime_utils.get_iso_date()
        now = datetime_utils.get_iso_date()

        self.open_checkpoint_payloads[name] = {
            "user_id": None,
            "job_id": shortuuid.uuid(),
            "session_id": self.session_id,
            "name": name,
            "long_name": name,
            "status": "running",
            "machine": self.machine,
            "date_created": start,
            "date_modified": now,
            "runtime": datetime_utils.get_runtime(start, now),
            "stdout_fn": None,
            "unread": True,
            "error_message": None,
        }
        if name not in self.checkpoints:
            self.checkpoints[name] = checkpoint.Checkpoint(name)
        logger = logging.getLogger(f"{name}.start")
        logger.debug("Profiling block started.")
        
    def stop(self, name=None, callbacks=[], to_db=False):
        """
        Close checkpoint and stop timer. Store start, stop, and elapsed times.
        Process job status payload and execute notification callbacks.

        If checkpoint name isn't supplied, get the most recent checkpoint (last
        in, first out; LIFO).

        Parameters
        ----------
        name : str, optional
            Checkpoint name
        callbacks : list, optional
            List of optional notification callbacks
        """
        if name is None:
            name = list(self.open_checkpoint_payloads.keys())[-1]
        elif name not in self.open_checkpoint_payloads:
            raise KeyError(f"No open checkpoint named {name}")

        payload = self.open_checkpoint_payloads.pop(name)
        if payload["status"] == "running":
            payload["status"] = "success"
        start = payload["date_created"]
        stop = datetime_utils._update_payload_times(payload)
        self.checkpoints[name].add_times(start, stop)

        logger = logging.getLogger(f"{name}.stop")
        logger.debug("Profiling block stopped.")
        formatted_runtime = checkpoint.format_reported_times(self.checkpoints[name].elapsed[-1])
        logger.info(f"Elapsed time: {formatted_runtime}")

        if self.to_db or to_db:
            with contextlib.closing(sqlite3.connect(f"{_config.JORT_DIR}/jort.db")) as con:
                cur = con.cursor()
                # Make sure session info is included in db
                sql = (
                    "INSERT OR IGNORE INTO sessions VALUES(?, ?)"
                )
                cur.execute(sql, (self.session_id, self.session_name))
                # Insert job into db
                sql = (
                    "INSERT INTO jobs VALUES("
                    "    :job_id,"
                    "    :session_id,"
                    "    :name,"
                    "    :status,"
                    "    :machine,"
                    "    :date_created,"
                    "    :date_modified,"
                    "    :runtime,"
                    "    :stdout_fn,"
                    "    :error_message"
                    ")"
                )
                cur.execute(sql, payload)
                job_id = cur.lastrowid
                con.commit()

        for callback in callbacks:
            callback.execute(payload=payload)
        
    def remove(self, name=None):
        """
        Option to remove checkpoint start instead of completing a profiling
        set, such as on catching an error.

        Parameters
        ----------
        name : str, optional
            Checkpoint name
        """
        if name is None:
            name = list(self.open_checkpoint_payloads.keys())[-1]
        
        if name in self.open_checkpoint_payloads:
            payload = self.open_checkpoint_payloads.pop(name)
            logger = logging.getLogger(f"{name}.remove")
            logger.debug("Profiling block removed.")
        
    def clear_open(self):
        """
        Clear all open checkpoints / open job status payloads.
        """
        self.open_checkpoint_payloads = {}

    def raise_error(self):
        """
        Update information payload with error details for the outermost checkpoint,
        to only be used within the except block during exception handling.
        """
        name = list(self.open_checkpoint_payloads.keys())[0]
        payload = self.open_checkpoint_payloads[name]
        payload["status"] = "error"
        payload["error_message"] = traceback.format_exc().strip().split('\n')[-1]
        raise

    def track(self, f=None, callbacks=[], report=False):
        """
        Function wrapper for tracker, to be used as a decorator. Creates a checkpoint
        with the input function's name. 

        Without parameters / evaluation, the decorator simply creates the checkpoint 
        and times the input function. With parameters, this method can execute 
        callbacks and print a report. 

        Parameters
        ----------
        f : func, optional
            Function to decorate
        callbacks : list, optional
            List of optional notification callbacks
        report : bool, optional
            Option to print tracker report at function completion
        """
        assert callable(f) or f is None
        def decorator(func):
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                self.start(name=func.__qualname__)
                try:
                    result = func(*args, **kwargs)
                except Exception as e:
                    payload = self.open_checkpoint_payloads[func.__qualname__]
                    payload["status"] = "error"
                    payload["error_message"] = traceback.format_exc().strip().split('\n')[-1]
                    raise
                finally:
                    self.stop(name=func.__qualname__, callbacks=callbacks)
                    if report:
                        self.report()
                return result
            return wrapper
        return decorator(f) if f else decorator
        
    def report(self, dec=1):
        """
        Print formatted runtime statistics for all checkpoints.

        Parameters
        ----------
        dec : int
            Decimal precision
        """
        for name in self.checkpoints:
            ckpt = self.checkpoints[name]
            print(ckpt.report(dec=dec))
            
            
def track(f=None, callbacks=[], report=True):
    """
    Independent function wrapper, to be used as a decorator, that creates a one-off
    tracker.
    
    Without parameters / evaluation, the decorator simply times the input function
    and prints a report by default. With parameters, this method can execute notification
    callbacks and control whether or not to print a report. 

    Parameters
    ----------
    f : func, optional
        Function to decorate
    callbacks : list, optional
        List of optional notification callbacks
    report : bool, optional
        Option to print tracker report at function completion
    """
    return Tracker(verbose=0).track(f=f, callbacks=callbacks, report=report)


from datetime import datetime
import dateutil.parser


def get_iso_date(timestamp=None):
    """
    Return start date in ISO 8601 format
    
    Parameters
    ----------
    timestamp : float, optional
        Unix time in seconds

    Returns
    -------
    iso_time : string
        ISO time (UTC)
    """
    if timestamp:
        return datetime.utcfromtimestamp(timestamp).isoformat()
    else:
        return datetime.utcnow().isoformat()

def get_runtime(iso_date1, iso_date2):
    """
    Return runtime in seconds between two dates in ISO format.

    Parameters
    ----------
    iso_date1 : string
        Start date in ISO 8601
    iso_date2 : string
        Stop date in ISO 8601

    Returns
    -------
    runtime : float
        Runtime in seconds
    """
    runtime = (
        dateutil.parser.parse(iso_date2) - dateutil.parser.parse(iso_date1)
    )
    return runtime.total_seconds()

def _update_payload_times(payload):
    """
    Modify payload dictionary with current time and time elapsed.

    Parameters
    ----------
    payload : dict
        Dictionary with job status details

    Returns
    -------
    date_now : string
        Current ISO date
    """
    date_now = get_iso_date()
    runtime = get_runtime(payload["date_created"], date_now)
    payload['runtime'] = runtime
    payload['date_modified'] = date_now
    return date_now


"""
Initialize jort directories with correct permissions.
"""

import os
import json
import sqlite3
from pathlib import Path


# Create internal jort directory
JORT_DIR = f"{os.path.expanduser('~')}/.jort"
Path(f"{JORT_DIR}/").mkdir(mode=0o700, parents=True, exist_ok=True)
Path(f"{JORT_DIR}/config").touch(mode=0o600, exist_ok=True)

def get_config_data():
    with open(f"{JORT_DIR}/config", "r") as f:
        try:
            config_data = json.load(f)
        except json.decoder.JSONDecodeError:
            config_data = {}
    return config_data

# Set up database
def _initialize_db():
    con = sqlite3.connect(f"{JORT_DIR}/jort.db")
    cur = con.cursor()

    sql = (
        "CREATE TABLE IF NOT EXISTS sessions ("
            "session_id TEXT PRIMARY KEY,"
            "session_name TEXT"
        ")"
    )
    cur.execute(sql)

    sql = (
        "CREATE TABLE IF NOT EXISTS jobs ("
        "    job_id TEXT PRIMARY KEY,"
        "    session_id TEXT,"
        "    job_name TEXT,"
        "    status TEXT,"
        "    machine TEXT,"
        "    date_created TEXT,"
        "    date_finished TEXT,"
        "    runtime REAL,"
        "    stdout_fn TEXT,"
        "    error_message TEXT,"
        "    FOREIGN KEY(session_id) REFERENCES sessions(session_id)"
        ")"
    )
    cur.execute(sql)

    con.commit()
    con.close()

_initialize_db()


from abc import ABC, abstractmethod
import os
import json
import smtplib
import ssl
import email
from email import encoders
from email.mime.base import MIMEBase
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart
from email.mime.application import MIMEApplication
import twilio.rest
import humanfriendly
from . import _config
from . import exceptions


class Callback(ABC):
    """
    Abstract base class for notification callbacks.
    """
    def __init__(self):
        pass

    @abstractmethod
    def format_message(self, payload):
        """
        Format notification message as a string.
        """
        pass

    @abstractmethod
    def execute(self, payload):
        """
        Send notification given job status payload.
        """
        pass
    
    
class PrintReport(Callback):
    """
    Print job runtime on completion.
    """
    def __init__(self):
        pass

    def format_message(self, payload):
        if payload["status"] == "success":
            return (
                f'\n'
                f'Your job `{payload["name"]}` successfully completed '
                f'in {humanfriendly.format_timespan(payload["runtime"])}'
            )
        elif payload["status"] == "error":
            error_text = payload["error_message"].split(":")[0]
            return (
                f'\n'
                f'Your job `{payload["name"]}` exited in error ({error_text}) '
                f'after {humanfriendly.format_timespan(payload["runtime"])}'
            )
        elif payload["status"] == "finished":
            return (
                f'\n'
                f'Your job `{payload["name"]}` finished running '
                f'in {humanfriendly.format_timespan(payload["runtime"])}'
            )
        else:
            raise exceptions.JortException(f'Invalid status: {payload["status"]}')

    def execute(self, payload):
        print(self.format_message(payload))


class EmailNotification(Callback):
    """
    Send email notifications to and from your email account. Requires login 
    credentials, which can be entered at the command line via :code:`jort -i`.
    """
    def __init__(self, email=None):
        config_data = _config.get_config_data()
        self.email = config_data.get("email")
        self.smtp_server = config_data.get("smtp_server")
        self.email_password = config_data.get("email_password")
        if email is not None:
            self.email = email

        if self.email_password is None:
            raise exceptions.JortCredentialException("Missing email password, add with `jort -i` command")
        if self.smtp_server is None:
            raise exceptions.JortException("Missing SMTP server, add with `jort -i` command")
        if self.email is None:
            raise exceptions.JortException("Missing email")

    def format_message(self, payload):
        if payload["machine"] is not None:
            machine_text = f' on machine {payload["machine"]}'
            html_machine_text = f' on machine <strong>{payload["machine"]}</strong>'
        else:
            machine_text = ''
            html_machine_text = ''

        if payload["status"] == "success":
            subject = "[jort] Your job finished successfully!"

            summary_text = (
                f'Your job `{payload["name"]}` completed at {payload["date_modified"]} (UTC)'
                f'{machine_text} with no errors.'
            )
            html_summary_text = (
                f'Your job <strong>{payload["name"]}</strong> completed at <strong>{payload["date_modified"]}</strong> (UTC)'
                f'{html_machine_text} with no errors.'
            )
        elif payload["status"] == "error":
            subject = "[jort] Your job exited with an error"

            error_text = payload["error_message"].split(":")[0]

            summary_text = (
                f'Your job `{payload["name"]}` exited at {payload["date_modified"]} (UTC)'
                f'{machine_text} with error `{error_text}`.'
            )
            html_summary_text = (
                f'Your job <strong>{payload["name"]}</strong> exited at <strong>{payload["date_modified"]}</strong> (UTC)'
                f'{html_machine_text} with error <strong>{error_text}</strong>.'
            )
        elif payload["status"] == "finished":
            subject = "[jort] Your job has finished!"

            summary_text = (
                f'Your job `{payload["name"]}` finished running at {payload["date_modified"]} (UTC)'
                f'{machine_text}.'
            )
            html_summary_text = (
                f'Your job <strong>{payload["name"]}</strong> finished running at <strong>{payload["date_modified"]}</strong> (UTC)'
                f'{html_machine_text}.'
            )
        else:
            raise exceptions.JortException(f'Invalid status: {payload["status"]}')
        
        runtime_text = f'The job\'s total runtime was {humanfriendly.format_timespan(payload["runtime"])}.'
        html_runtime_text = f'The job\'s total runtime was <strong>{humanfriendly.format_timespan(payload["runtime"])}</strong>.'

        body = (
            f'{summary_text}\r\n'
            f'{runtime_text}\r\n'
            f'--\r\n'
            f'jort'
        )
        html_body = (
            f'<html>'
            f'<head></head>'
            f'<body>'
            f'  <p>{html_summary_text}</p>'
            f'  <p>{html_runtime_text}</p>'
            f'  <p>--<br>jort</p>'
            f'</body>'
            f'</html>'
        )
        email_data = {
            "subject": subject,
            "body": body,
            "html_body": html_body,
        }
        return email_data

    def execute(self, payload):
        email_data = self.format_message(payload)

        message = MIMEMultipart("alternative")
        message.attach(MIMEText(email_data["body"], "plain"))
        message.attach(MIMEText(email_data["html_body"], "html"))

        if payload["stdout_fn"] is not None:
            stdout_path = f'{_config.JORT_DIR}/{payload["stdout_fn"]}'
            with open(stdout_path, "r") as f:
                attachment = MIMEApplication(f.read(), _subtype="txt")
            attachment.add_header("Content-Disposition", "attachment", filename="output.txt")

            message_mix = MIMEMultipart("mixed")
            message_mix.attach(message)
            message_mix.attach(attachment)
            message = message_mix

        message["Subject"] = email_data["subject"]
        message["From"] = self.email
        message["To"] = self.email

        # Secure connection
        context = ssl.create_default_context()
        with smtplib.SMTP_SSL(self.smtp_server, port=465, context=context) as server:
            server.login(self.email, self.email_password)
            server.sendmail(message["From"], message["To"], message.as_string())


class TextNotification(Callback):
    """
    Send SMS notifications to and from numbers managed by your Twilio account. Requires 
    Twilio credentials, which can be entered at the command line via :code:`jort -i`.
    """
    def __init__(self, receive_number=None):
        config_data = _config.get_config_data()
        self.receive_number = config_data.get("twilio_receive_number")
        self.send_number = config_data.get("twilio_send_number")
        self.twilio_account_sid = config_data.get("twilio_account_sid")
        self.twilio_auth_token = config_data.get("twilio_auth_token")
        if receive_number is not None:
            self.receive_number = receive_number

        if self.twilio_account_sid is None or self.twilio_auth_token is None:
            raise exceptions.JortCredentialException("Missing Twilio credentials, add with `jort -i` command")
        if self.send_number is None:
            raise exceptions.JortException("Missing Twilio sending number, add with `jort -i` command")
        if self.receive_number is None:
            raise exceptions.JortException("Missing receiving number")

    def format_message(self, payload):
        if payload["status"] == "success":
            return (
                f'Your job `{payload["name"]}` successfully completed '
                f'in {humanfriendly.format_timespan(payload["runtime"])}'
            )
        elif payload["status"] == "error":
            error_text = payload["error_message"].split(":")[0]
            return (
                f'Your job `{payload["name"]}` exited in error ({error_text}) '
                f'after {humanfriendly.format_timespan(payload["runtime"])}'
            )
        elif payload["status"] == "finished":
            return (
                f'Your job `{payload["name"]}` finished running '
                f'in {humanfriendly.format_timespan(payload["runtime"])}'
            )
        else:
            raise exceptions.JortException(f'Invalid status: {payload["status"]}')
    
    def execute(self, payload):
        client = twilio.rest.Client(self.twilio_account_sid,
                                    self.twilio_auth_token)
        message = client.messages.create(body=self.format_message(payload),
                                         from_=self.send_number,
                                         to=self.receive_number)





import numpy as np
from . import datetime_utils


class Checkpoint(object):
    """
    Checkpoint object to mark start, stop, and elapsed times.

    Start and stop times are in ISO format, while elapsed times are in seconds.

    Parameters
    ----------
    name : string
        Checkpoint name

    :ivar name: checkpoint name
    :ivar starts: list of start times
    :ivar stops: list of stop times
    :ivar elapsed: list of elapsed times
    """
    def __init__(self, name):
        self.name = name 
        self.starts = [] 
        self.stops = [] 
        self.elapsed = [] 
        
    def _get_elapsed(self):
        self.elapsed = [datetime_utils.get_runtime(t1, t2) 
                        for (t1, t2) in zip(self.starts, self.stops)]
        
    def add_times(self, start, stop):
        """
        Store start, stop, and elpased times to checkpoint.

        Parameters
        ----------
        start : string
            ISO date of start
        stop : string
            ISO date of stop
        """
        self.starts.append(start)
        self.stops.append(stop)
        self.elapsed.append(datetime_utils.get_runtime(start, stop))
    
    def report(self, dec=1):
        """
        Return formatted runtime statistics.

        Parameters
        ----------
        dec : int
            Decimal precision

        Returns
        -------
        format_str : string
            Runtime details
        """
        if len(self.elapsed) > 0:
            elapsed_stats = format_reported_times(np.mean(self.elapsed), 
                                                np.std(self.elapsed), 
                                                dec=dec)
            return (
                f"{self.name} | "
                f"{elapsed_stats} per iteration, "
                f"n = {len(self.elapsed)}"
            )
        else:
            return (
                f"{self.name} | "
                f"no complete iterations, "
                f"n = {len(self.elapsed)}"
            )
        
    def __str__(self):
        return self.report(dec=1)
    

def format_reported_times(m, s=None, dec=1):
    """
    Format runtime strings, automatically adjusting units if runtimes 
    are under a second. 

    Parameters
    ----------
    m : float
        Mean runtime
    s : float, optional
        Standard deviation runtime
    dec : int, optional
        Decimal precision

    Returns
    -------
    format_str : string
        Runtime as a string
    """
    format_str = ""
    if m >= 1.:
        format_str += f"{m:.{dec}f} s" 
        if s is not None:
            format_str += f" \u00B1 {s:.{dec}f} s"
    elif m >= 1e-3:
        format_str += f"{1e3 * m:.{dec}f} ms" 
        if s is not None:
            format_str += f" \u00B1 {1e3 * s:.{dec}f} ms"
    elif m >= 1e-6:
        format_str += f"{1e6 * m:.{dec}f} us" 
        if s is not None:
            format_str += f" \u00B1 {1e6 * s:.{dec}f} us"
    else:
        format_str += f"{1e9 * m:.{dec}f} ns" 
        if s is not None:
            format_str += f" \u00B1 {1e9 * s:.{dec}f} ns"
    return format_str


from __future__ import absolute_import, division, print_function

from ._version import __version__

from .tracker import Tracker, track
from .track_cli import track_new, track_existing
from .reporting_callbacks import EmailNotification, TextNotification, PrintReport
from .exceptions import JortException, JortCredentialException



import os
import sys
import time
import sqlite3
import contextlib
import shlex
import subprocess
import uuid
import psutil
import shortuuid
from tqdm import tqdm
from pprint import pprint

from . import tracker
from . import datetime_utils
from . import _config
from . import reporting_callbacks


def track_new(command,
              use_shell=False,
              store_stdout=False,
              save_filename=None,
              to_db=False,
              session_name=None,
              unique=False,
              send_text=False,
              send_email=False,
              verbose=False,
              update_period=-1):
    """
    Track execution time and details of new command line process.

    Parameters
    ----------
    command : str
        Command to execute, which is spawned as a subprocess
    use_shell : bool, optional
        Option to use shell execution for subprocess
    store_stdout : bool, optional
        Option to write command output to file
    save_filename : str, optional
        Filename to which to save command output
    to_db : bool, optional
        Save all checkpoints to database
    session_name : str, optional
        Name of job session, if saving jobs to database
    unique : bool, optional
        Whether to skip job, if already successfully run and stored in database
    send_text : bool, optional
        Option to send SMS notification on completion
    send_email : bool, optional
        Option to send e-mail notification on completion
    verbose : bool, optional
        Option to control how much information is printed in stdout
    update_period : int, optional
        Number of seconds between each payload update and stdout write. If 
        :code:`update_period=-1`, as default, the only update occurs on completion.
    """
    callbacks = [reporting_callbacks.PrintReport()]
    if send_text:
        callbacks.append(reporting_callbacks.TextNotification())
    if send_email:
        callbacks.append(reporting_callbacks.EmailNotification())

    # Key for storing stdout text to file
    if save_filename or store_stdout:
        stdout_fn = f"{shortuuid.uuid()}.txt"
        stdout_path = f"{_config.JORT_DIR}/{stdout_fn}"
    else:
        stdout_fn = None

    tr = tracker.Tracker(to_db=to_db, session_name=session_name)
    if unique:
        with contextlib.closing(sqlite3.connect(f"{_config.JORT_DIR}/jort.db")) as con:
            cur = con.cursor()
            sql = (
                "SELECT status FROM jobs WHERE session_id == ? AND job_name == ?"
            )
            res = cur.execute(sql, (tr.session_id, command))
            for row in res.fetchall():
                status = row[0]
                if status == "success":
                    print("Found matching job that completed successfully; skipping...")
                    con.close()
                    return

    tr.start(name=command)

    payload = tr.open_checkpoint_payloads[command]

    payload['stdout_fn'] = stdout_fn

    # ACTUALLY START SUBPROCESS
    my_env = os.environ.copy()
    my_env["PYTHONUNBUFFERED"] = "1"

    if use_shell:
        p = psutil.Popen(command,
                         shell=True,
                         env=my_env,
                         stdout=subprocess.PIPE,
                         stderr=subprocess.STDOUT,
                         bufsize=1,
                         universal_newlines=True)
    else:
        p = psutil.Popen(shlex.split(command),
                         shell=False,
                         env=my_env,
                         stdout=subprocess.PIPE,
                         stderr=subprocess.STDOUT,
                         bufsize=1,
                         universal_newlines=True)
    print(f"Subprocess PID: {p.pid}\n")

    # Create stdout file
    if verbose:
        pprint(payload)
    if save_filename or store_stdout:
        with open(stdout_path, "a+") as f:
            f.write(f"{command}\n")
            f.write(f"--\n")

    buffer = ""
    temp_start = time.time()
    for line in p.stdout:
        if update_period > 0 and time.time() - temp_start >= update_period:
            if verbose:
                print("Buffered! (Not sent)", [buffer])
            if save_filename or store_stdout:
                with open(stdout_path, "a+") as f:
                    f.write(buffer)

            payload['status'] = 'running'
            datetime_utils._update_payload_times(payload)
            if verbose:
                pprint(payload)

            buffer = ""
            temp_start = time.time()

        sys.stdout.write(line)
        buffer += line


    if verbose:
        print("Buffered!", [buffer])
    if save_filename or store_stdout:
        with open(stdout_path, "a+") as f:
            f.write(buffer)

    p.wait()

    if verbose:
        print(f"Exit code: {p.returncode}")

    if p.returncode == 0:
        payload["status"] = "success"
    else:
        payload["status"] = "error"
        payload["error_message"] = line
    tr.stop(callbacks=callbacks)
    # print("")
    # if payload["runtime"] < 10:
    #     sys.exit("Job exited in 10 seconds -- no need to track!")

    if verbose:
        pprint(payload)

    if save_filename or store_stdout:
        if save_filename:
            subprocess.call(["cp", stdout_path, save_filename])
        try:
            subprocess.call(["rm", stdout_path])
        except Exception as e:
            raise e


def track_existing(pid,
                   to_db=False,
                   session_name=None,
                   send_text=False,
                   send_email=False,
                   verbose=False,
                   update_period=-1):
    """
    Track execution time and details of existing command line process.

    Parameters
    ----------
    pid : int
        Process ID of existing process
    to_db : bool, optional
        Save all checkpoints to database
    session_name : str, optional
        Name of job session, if saving jobs to database
    send_text : bool, optional
        Option to send SMS notification on completion
    send_email : bool, optional
        Option to send e-mail notification on completion
    verbose : bool, optional
        Option to control how much information is printed in stdout
    update_period : int, optional
        Number of seconds between each payload update. If 
        :code:`update_period=-1`, as default, the only update occurs on completion.
    """
    callbacks = [reporting_callbacks.PrintReport()]
    if send_text:
        callbacks.append(reporting_callbacks.TextNotification())
    if send_email:
        callbacks.append(reporting_callbacks.EmailNotification())

    # Does not support stdout tracking
    stdout_fn = None

    # Create process based on PID and grab relevant information
    p = psutil.Process(pid)
    command = " ".join(p.cmdline())

    tr = tracker.Tracker(to_db=to_db, session_name=session_name)
    tr.start(name=command, date_created=datetime_utils.get_iso_date(p.create_time()))
    payload = tr.open_checkpoint_payloads[command]

    if verbose:
        pprint(payload)

    temp_start = time.time()

    while p.is_running():
        if update_period > 0 and time.time() - temp_start >= update_period:
            payload["status"] = "running"
            datetime_utils._update_payload_times(payload)
            if verbose:
                pprint(payload)

            temp_start = time.time()

    payload["status"] = "finished"
    tr.stop(callbacks=callbacks)

    # print("")
    # if runtime_s < 60:
    #     sys.exit("Job exited in less than a minute -- no need to track!")

    if verbose:
        pprint(payload)